@@ -32,50 +32,27 @@ use serde::Deserialize;
 use serde_json;
 
 use super::core::*;
+use super::delete::GcsDeleter;
 use super::error::parse_error;
 use super::lister::GcsLister;
 use super::writer::GcsWriter;
 use super::writer::GcsWriters;
+use crate::raw::oio::BatchDeleter;
 use crate::raw::*;
+use crate::services::GcsConfig;
 use crate::*;
 
 const DEFAULT_GCS_ENDPOINT: &str = "https://storage.googleapis.com";
 const DEFAULT_GCS_SCOPE: &str = "https://www.googleapis.com/auth/devstorage.read_write";
 
-/// [Google Cloud Storage](https://cloud.google.com/storage) services support.
-#[derive(Default, Deserialize)]
-#[serde(default)]
-#[non_exhaustive]
-pub struct GcsConfig {
-    /// root URI, all operations happens under `root`
-    root: Option<String>,
-    /// bucket name
-    bucket: String,
-    /// endpoint URI of GCS service,
-    /// default is `https://storage.googleapis.com`
-    endpoint: Option<String>,
-    /// Scope for gcs.
-    scope: Option<String>,
-    /// Service Account for gcs.
-    service_account: Option<String>,
-    /// Credentials string for GCS service OAuth2 authentication.
-    credential: Option<String>,
-    /// Local path to credentials file for GCS service OAuth2 authentication.
-    credential_path: Option<String>,
-    /// The predefined acl for GCS.
-    predefined_acl: Option<String>,
-    /// The default storage class used by gcs.
-    default_storage_class: Option<String>,
-}
-
-impl Debug for GcsConfig {
-    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
-        f.debug_struct("GcsConfig")
-            .field("root", &self.root)
-            .field("bucket", &self.bucket)
-            .field("endpoint", &self.endpoint)
-            .field("scope", &self.scope)
-            .finish_non_exhaustive()
+impl Configurator for GcsConfig {
+    type Builder = GcsBuilder;
+    fn into_builder(self) -> Self::Builder {
+        GcsBuilder {
+            config: self,
+            http_client: None,
+            customized_token_loader: None,
+        }
     }
 }
 
@@ -86,7 +63,7 @@ pub struct GcsBuilder {
     config: GcsConfig,
 
     http_client: Option<HttpClient>,
-    customed_token_loader: Option<Box<dyn GoogleTokenLoad>>,
+    customized_token_loader: Option<Box<dyn GoogleTokenLoad>>,
 }
 
 impl Debug for GcsBuilder {
@@ -100,16 +77,18 @@ impl Debug for GcsBuilder {
 
 impl GcsBuilder {
     /// set the working directory root of backend
-    pub fn root(&mut self, root: &str) -> &mut Self {
-        if !root.is_empty() {
-            self.config.root = Some(root.to_string())
-        }
+    pub fn root(mut self, root: &str) -> Self {
+        self.config.root = if root.is_empty() {
+            None
+        } else {
+            Some(root.to_string())
+        };
 
         self
     }
 
     /// set the container's name
-    pub fn bucket(&mut self, bucket: &str) -> &mut Self {
+    pub fn bucket(mut self, bucket: &str) -> Self {
         self.config.bucket = bucket.to_string();
         self
     }
@@ -125,7 +104,7 @@ impl GcsBuilder {
     /// - full-control: `https://www.googleapis.com/auth/devstorage.full_control`
     ///
     /// Reference: [Cloud Storage authentication](https://cloud.google.com/storage/docs/authentication)
-    pub fn scope(&mut self, scope: &str) -> &mut Self {
+    pub fn scope(mut self, scope: &str) -> Self {
         if !scope.is_empty() {
             self.config.scope = Some(scope.to_string())
         };
@@ -136,7 +115,7 @@ impl GcsBuilder {
     ///
     /// service account will be used for fetch token from vm metadata.
     /// If not set, we will try to fetch with `default` service account.
-    pub fn service_account(&mut self, service_account: &str) -> &mut Self {
+    pub fn service_account(mut self, service_account: &str) -> Self {
         if !service_account.is_empty() {
             self.config.service_account = Some(service_account.to_string())
         };
@@ -144,7 +123,7 @@ impl GcsBuilder {
     }
 
     /// set the endpoint GCS service uses
-    pub fn endpoint(&mut self, endpoint: &str) -> &mut Self {
+    pub fn endpoint(mut self, endpoint: &str) -> Self {
         if !endpoint.is_empty() {
             self.config.endpoint = Some(endpoint.to_string())
         };
@@ -158,7 +137,7 @@ impl GcsBuilder {
     /// we will use one of `credential` and `credential_path` to complete the OAuth2 authentication.
     ///
     /// Reference: [Google Cloud Storage Authentication](https://cloud.google.com/docs/authentication).
-    pub fn credential(&mut self, credential: &str) -> &mut Self {
+    pub fn credential(mut self, credential: &str) -> Self {
         if !credential.is_empty() {
             self.config.credential = Some(credential.to_string())
         };
@@ -171,7 +150,7 @@ impl GcsBuilder {
     /// we will use one of `credential` and `credential_path` to complete the OAuth2 authentication.
     ///
     /// Reference: [Google Cloud Storage Authentication](https://cloud.google.com/docs/authentication).
-    pub fn credential_path(&mut self, path: &str) -> &mut Self {
+    pub fn credential_path(mut self, path: &str) -> Self {
         if !path.is_empty() {
             self.config.credential_path = Some(path.to_string())
         };
@@ -184,14 +163,32 @@ impl GcsBuilder {
     ///
     /// This API is part of OpenDAL's Raw API. `HttpClient` could be changed
     /// during minor updates.
-    pub fn http_client(&mut self, client: HttpClient) -> &mut Self {
+    pub fn http_client(mut self, client: HttpClient) -> Self {
         self.http_client = Some(client);
         self
     }
 
-    /// Specify the customed token loader used by this service.
-    pub fn customed_token_loader(&mut self, token_load: Box<dyn GoogleTokenLoad>) -> &mut Self {
-        self.customed_token_loader = Some(token_load);
+    /// Specify the customized token loader used by this service.
+    pub fn customized_token_loader(mut self, token_load: Box<dyn GoogleTokenLoad>) -> Self {
+        self.customized_token_loader = Some(token_load);
+        self
+    }
+
+    /// Provide the OAuth2 token to use.
+    pub fn token(mut self, token: String) -> Self {
+        self.config.token = Some(token);
+        self
+    }
+
+    /// Disable attempting to load credentials from the GCE metadata server.
+    pub fn disable_vm_metadata(mut self) -> Self {
+        self.config.disable_vm_metadata = true;
+        self
+    }
+
+    /// Disable loading configuration from the environment.
+    pub fn disable_config_load(mut self) -> Self {
+        self.config.disable_config_load = true;
         self
     }
 
@@ -204,7 +201,7 @@ impl GcsBuilder {
     /// - `private`
     /// - `projectPrivate`
     /// - `publicRead`
-    pub fn predefined_acl(&mut self, acl: &str) -> &mut Self {
+    pub fn predefined_acl(mut self, acl: &str) -> Self {
         if !acl.is_empty() {
             self.config.predefined_acl = Some(acl.to_string())
         };
@@ -218,32 +215,31 @@ impl GcsBuilder {
     /// - `NEARLINE`
     /// - `COLDLINE`
     /// - `ARCHIVE`
-    pub fn default_storage_class(&mut self, class: &str) -> &mut Self {
+    pub fn default_storage_class(mut self, class: &str) -> Self {
         if !class.is_empty() {
             self.config.default_storage_class = Some(class.to_string())
         };
         self
     }
+
+    /// Allow anonymous requests.
+    ///
+    /// This is typically used for buckets which are open to the public or GCS
+    /// storage emulators.
+    pub fn allow_anonymous(mut self) -> Self {
+        self.config.allow_anonymous = true;
+        self
+    }
 }
 
 impl Builder for GcsBuilder {
     const SCHEME: Scheme = Scheme::Gcs;
-    type Accessor = GcsBackend;
+    type Config = GcsConfig;
 
-    fn from_map(map: HashMap<String, String>) -> Self {
-        let config = GcsConfig::deserialize(ConfigDeserializer::new(map))
-            .expect("config deserialize must succeed");
-
-        GcsBuilder {
-            config,
-            ..GcsBuilder::default()
-        }
-    }
-
-    fn build(&mut self) -> Result<Self::Accessor> {
+    fn build(self) -> Result<impl Access> {
         debug!("backend build started: {:?}", self);
 
-        let root = normalize_root(&self.config.root.take().unwrap_or_default());
+        let root = normalize_root(&self.config.root.unwrap_or_default());
         debug!("backend use root {}", root);
 
         // Handle endpoint and bucket name
@@ -258,7 +254,7 @@ impl Builder for GcsBuilder {
 
         // TODO: server side encryption
 
-        let client = if let Some(client) = self.http_client.take() {
+        let client = if let Some(client) = self.http_client {
             client
         } else {
             HttpClient::new().map_err(|err| {
@@ -287,21 +283,31 @@ impl Builder for GcsBuilder {
             cred_loader = cred_loader.with_disable_well_known_location();
         }
 
+        if self.config.disable_config_load {
+            cred_loader = cred_loader
+                .with_disable_env()
+                .with_disable_well_known_location();
+        }
+
         let scope = if let Some(scope) = &self.config.scope {
             scope
         } else {
             DEFAULT_GCS_SCOPE
         };
 
-        let mut token_loader = GoogleTokenLoader::new(scope, client.client());
+        let mut token_loader = GoogleTokenLoader::new(scope, GLOBAL_REQWEST_CLIENT.clone());
         if let Some(account) = &self.config.service_account {
             token_loader = token_loader.with_service_account(account);
         }
         if let Ok(Some(cred)) = cred_loader.load() {
             token_loader = token_loader.with_credentials(cred)
         }
-        if let Some(loader) = self.customed_token_loader.take() {
-            token_loader = token_loader.with_customed_token_loader(loader)
+        if let Some(loader) = self.customized_token_loader {
+            token_loader = token_loader.with_customized_token_loader(loader)
+        }
+
+        if self.config.disable_vm_metadata {
+            token_loader = token_loader.with_disable_vm_metadata(true);
         }
 
         let signer = GoogleSigner::new("storage");
@@ -314,9 +320,12 @@ impl Builder for GcsBuilder {
                 client,
                 signer,
                 token_loader,
+                token: self.config.token,
+                scope: scope.to_string(),
                 credential_loader: cred_loader,
                 predefined_acl: self.config.predefined_acl.clone(),
                 default_storage_class: self.config.default_storage_class.clone(),
+                allow_anonymous: self.config.allow_anonymous,
             }),
         };
 
@@ -334,11 +343,13 @@ impl Access for GcsBackend {
     type Reader = HttpBody;
     type Writer = GcsWriters;
     type Lister = oio::PageLister<GcsLister>;
+    type Deleter = oio::BatchDeleter<GcsDeleter>;
     type BlockingReader = ();
     type BlockingWriter = ();
     type BlockingLister = ();
+    type BlockingDeleter = ();
 
-    fn info(&self) -> AccessorInfo {
+    fn info(&self) -> Arc<AccessorInfo> {
         let mut am = AccessorInfo::default();
         am.set_scheme(Scheme::Gcs)
             .set_root(&self.core.root)
@@ -347,6 +358,12 @@ impl Access for GcsBackend {
                 stat: true,
                 stat_with_if_match: true,
                 stat_with_if_none_match: true,
+                stat_has_etag: true,
+                stat_has_content_md5: true,
+                stat_has_content_length: true,
+                stat_has_content_type: true,
+                stat_has_last_modified: true,
+                stat_has_user_metadata: true,
 
                 read: true,
 
@@ -357,6 +374,9 @@ impl Access for GcsBackend {
                 write_can_empty: true,
                 write_can_multi: true,
                 write_with_content_type: true,
+                write_with_user_metadata: true,
+                write_with_if_not_exists: true,
+
                 // The min multipart size of Gcs is 5 MiB.
                 //
                 // ref: <https://cloud.google.com/storage/docs/xml-api/put-object-multipart>
@@ -371,23 +391,29 @@ impl Access for GcsBackend {
                 },
 
                 delete: true,
+                delete_max_size: Some(100),
                 copy: true,
 
                 list: true,
                 list_with_limit: true,
                 list_with_start_after: true,
                 list_with_recursive: true,
+                list_has_etag: true,
+                list_has_content_md5: true,
+                list_has_content_length: true,
+                list_has_content_type: true,
+                list_has_last_modified: true,
 
-                batch: true,
-                batch_max_operations: Some(100),
                 presign: true,
                 presign_stat: true,
                 presign_read: true,
                 presign_write: true,
 
+                shared: true,
+
                 ..Default::default()
             });
-        am
+        am.into()
     }
 
     async fn stat(&self, path: &str, args: OpStat) -> Result<RpStat> {
@@ -418,6 +444,10 @@ impl Access for GcsBackend {
 
         m.set_last_modified(parse_datetime_from_rfc3339(&meta.updated)?);
 
+        if !meta.metadata.is_empty() {
+            m.with_user_metadata(meta.metadata);
+        }
+
         Ok(RpStat::new(m))
     }
 
@@ -447,15 +477,11 @@ impl Access for GcsBackend {
         Ok((RpWrite::default(), w))
     }
 
-    async fn delete(&self, path: &str, _: OpDelete) -> Result<RpDelete> {
-        let resp = self.core.gcs_delete_object(path).await?;
-
-        // deleting not existing objects is ok
-        if resp.status().is_success() || resp.status() == StatusCode::NOT_FOUND {
-            Ok(RpDelete::default())
-        } else {
-            Err(parse_error(resp))
-        }
+    async fn delete(&self) -> Result<(RpDelete, Self::Deleter)> {
+        Ok((
+            RpDelete::default(),
+            BatchDeleter::new(GcsDeleter::new(self.core.clone())),
+        ))
     }
 
     async fn list(&self, path: &str, args: OpList) -> Result<(RpList, Self::Lister)> {
@@ -491,7 +517,7 @@ impl Access for GcsBackend {
             }
         };
 
-        self.core.sign_query(&mut req, args.expire()).await?;
+        self.core.sign_query(&mut req, args.expire())?;
 
         // We don't need this request anymore, consume it directly.
         let (parts, _) = req.into_parts();
@@ -502,65 +528,6 @@ impl Access for GcsBackend {
             parts.headers,
         )))
     }
-
-    async fn batch(&self, args: OpBatch) -> Result<RpBatch> {
-        let ops = args.into_operation();
-        if ops.len() > 100 {
-            return Err(Error::new(
-                ErrorKind::Unsupported,
-                "gcs services only allow delete less than 100 keys at once",
-            )
-            .with_context("length", ops.len().to_string()));
-        }
-
-        let paths: Vec<String> = ops.into_iter().map(|(p, _)| p).collect();
-        let resp = self.core.gcs_delete_objects(paths.clone()).await?;
-
-        let status = resp.status();
-
-        if let StatusCode::OK = status {
-            let content_type = parse_content_type(resp.headers())?.ok_or_else(|| {
-                Error::new(
-                    ErrorKind::Unexpected,
-                    "gcs batch delete response content type is empty",
-                )
-            })?;
-            let boundary = content_type
-                .strip_prefix("multipart/mixed; boundary=")
-                .ok_or_else(|| {
-                    Error::new(
-                        ErrorKind::Unexpected,
-                        "gcs batch delete response content type is not multipart/mixed",
-                    )
-                })?
-                .trim_matches('"');
-            let multipart: Multipart<MixedPart> = Multipart::new()
-                .with_boundary(boundary)
-                .parse(resp.into_body().to_bytes())?;
-            let parts = multipart.into_parts();
-
-            let mut batched_result = Vec::with_capacity(parts.len());
-
-            for (i, part) in parts.into_iter().enumerate() {
-                let resp = part.into_response();
-                // TODO: maybe we can take it directly?
-                let path = paths[i].clone();
-
-                // deleting not existing objects is ok
-                if resp.status().is_success() || resp.status() == StatusCode::NOT_FOUND {
-                    batched_result.push((path, Ok(RpDelete::default().into())));
-                } else {
-                    batched_result.push((path, Err(parse_error(resp))));
-                }
-            }
-
-            Ok(RpBatch::new(batched_result))
-        } else {
-            // If the overall request isn't formatted correctly and Cloud Storage is unable to parse it into sub-requests, you receive a 400 error.
-            // Otherwise, Cloud Storage returns a 200 status code, even if some or all of the sub-requests fail.
-            Err(parse_error(resp))
-        }
-    }
 }
 
 /// The raw json response returned by [`get`](https://cloud.google.com/storage/docs/json_api/v1/objects/get)
@@ -587,6 +554,10 @@ struct GetObjectJsonResponse {
     ///
     /// For example: `"contentType": "image/png",`
     content_type: String,
+    /// Custom metadata of this object.
+    ///
+    /// For example: `"metadata" : { "my-key": "my-value" }`
+    metadata: HashMap<String, String>,
 }
 
 #[cfg(test)]
@@ -612,7 +583,10 @@ mod tests {
   "etag": "CKWasoTgyPkCEAE=",
   "timeCreated": "2022-08-15T11:33:34.866Z",
   "updated": "2022-08-15T11:33:34.866Z",
-  "timeStorageClassUpdated": "2022-08-15T11:33:34.866Z"
+  "timeStorageClassUpdated": "2022-08-15T11:33:34.866Z",
+  "metadata" : {
+    "location" : "everywhere"
+  }
 }"#;
 
         let meta: GetObjectJsonResponse =
@@ -623,5 +597,9 @@ mod tests {
         assert_eq!(meta.md5_hash, "fHcEH1vPwA6eTPqxuasXcg==");
         assert_eq!(meta.etag, "CKWasoTgyPkCEAE=");
         assert_eq!(meta.content_type, "image/png");
+        assert_eq!(
+            meta.metadata,
+            HashMap::from_iter([("location".to_string(), "everywhere".to_string())])
+        );
     }
 }
